@@ -3,6 +3,7 @@ Python API for CSR matrices.
 """
 
 import warnings
+import logging
 import numpy as np
 import scipy.sparse as sps
 
@@ -13,15 +14,16 @@ from csr.kernels import get_kernel, releasing
 from . import _struct, _rows
 
 INTC = np.iinfo(np.intc)
+_log = logging.getLogger(__name__)
 
 # ugly hack for a bug on Numba < 0.53
 if config.DISABLE_JIT:
     class _csr_base:
-        def __init__(self, nrows, ncols, nnz, ptrs, inds, vals, cast=True):
+        def __init__(self, nrows, ncols, nnz, ptrs, inds, vals, _cast=True):
             self.nrows = nrows
             self.ncols = ncols
             self.nnz = nnz
-            if cast and np.max(ptrs, initial=0) <= INTC.max:
+            if _cast and np.max(ptrs, initial=0) <= INTC.max:
                 self.rowptrs = np.require(ptrs, np.intc, 'C')
             else:
                 self.rowptrs = np.require(ptrs, requirements='C')
@@ -51,7 +53,8 @@ class CSR(_csr_base):
     * It is usable from code compiled in Numba's nopython mode.
 
     You generally don't want to create this class yourself with the constructor.  Instead, use one
-    of its class or static methods.
+    of its class or static methods.  If you do use the constructor, be advised that the class may
+    reuse the arrays that you pass, but does not guarantee that they will be used.
 
     Not all methods are available from Numba, and a few have restricted signatures.  The
     documentation for each method notes deviations when in Numba-compiled code.
@@ -73,17 +76,21 @@ class CSR(_csr_base):
         values(numpy.ndarray or None): the values.
     """
 
-    def __new__(cls, nrows, ncols, nnz, rps, cis, vs, cast=True):
-        assert nrows >= INTC.min
+    def __new__(cls, nrows, ncols, nnz, rps, cis, vs, _cast=True):
+        assert nrows >= 0
         assert nrows <= INTC.max
-        assert ncols >= INTC.min
+        assert ncols >= 0
         assert ncols <= INTC.max
+        assert nnz >= 0
         nrows = np.intc(nrows)
         ncols = np.intc(ncols)
-        if cast:
+
+        if _cast:
             cis = np.require(cis, np.intc, 'C')
             if nnz <= INTC.max:
                 rps = np.require(rps, np.intc, 'C')
+            else:
+                rps = np.require(rps, np.int64, 'C')
             if vs is not None:
                 vs = np.require(vs, requirements='C')
 
@@ -92,7 +99,6 @@ class CSR(_csr_base):
         else:
             return _csr_base.__new__(cls)
 
-
     @classmethod
     def empty(cls, nrows, ncols, row_nnzs=None, values=True):
         """
@@ -103,15 +109,30 @@ class CSR(_csr_base):
             ncols(int): the number of columns.
             row_nnzs(array-like):
                 the number of nonzero entries for each row, or None for an empty matrix.
-            values(bool):
-                whether it has values or only structure.
+            values(bool, str, or numpy.dtype):
+                whether it has values or only structure; can be a NumPy data type to
+                specify a type other than `f8`.
         """
-        from .constructors import create_empty, create_from_sizes
+        from .constructors import create_empty
+        assert nrows >= 0
+        assert ncols >= 0
         if row_nnzs is not None:
             assert len(row_nnzs) == nrows
-            if nrows > 0 and np.max(row_nnzs) <= INTC.max:
-                row_nnzs = np.require(row_nnzs, np.intc)
-            return create_from_sizes(nrows, ncols, row_nnzs)
+            nnz = np.sum(row_nnzs, dtype=np.int64)
+            assert nnz >= 0
+            rp_dtype = np.intc if nnz <= INTC.max else np.int64
+            rps = np.zeros(nrows + 1, dtype=rp_dtype)
+            np.cumsum(row_nnzs, dtype=rp_dtype, out=rps[1:])
+            cis = np.zeros(nnz, dtype=np.int32)
+
+            if values is True:
+                vs = np.zeros(nnz)
+            elif values:
+                vs = np.zeros(nnz, dtype=values)
+            else:
+                vs = None
+
+            return cls(nrows, ncols, nnz, rps, cis, vs)
         else:
             return create_empty(nrows, ncols)
 
@@ -226,12 +247,52 @@ class CSR(_csr_base):
             return vs
 
     def _e_value(self, i):
+        """
+        Get the value of a particular element, returning 1 if values is undefined.
+        """
         vs = self.values
         if vs is not None:
             return vs[i]
         else:
             return 1.0
 
+    def _normalize(self, val_dtype=np.float64, ptr_dtype=None):
+        """
+        Normalize the matrix into a predictable structure and type.  It avoids copying
+        if possible.
+
+        .. note:: This method is not available from Numba.
+
+        Args:
+            val_dtype(np.dtype or None or boolean):
+                The value data type.  If ``False``, drop the value array.  If ``None``,
+                leave unchanged.
+            ptr_dtype(np.dtype or None):
+                The row pointer data type.  If ``None``, leave rows untransformed.
+        Returns:
+            CSR: the transformed CSR matrix.
+        """
+
+        if ptr_dtype:
+            info = np.iinfo(ptr_dtype)
+            if self.nnz > info.max:
+                raise ValueError(f'type {ptr_dtype} cannot address {self.nnz} entries')
+            rps = np.require(self.rowptrs, ptr_dtype)
+        else:
+            rps = self.rowptrs
+
+        if val_dtype:
+            if self.values is None:
+                vs = np.ones(self.nnz, val_dtype)
+            else:
+                vs = np.require(self.values, val_dtype)
+        elif val_dtype is False:
+            vs = None
+        else:
+            vs = self.values
+
+        return CSR(self.nrows, self.ncols, self.nnz, rps, self.colinds, vs, _cast=False)
+
     @property
     def R(self):
         warnings.warn('.R deprecated, use CSR directly', DeprecationWarning)
@@ -462,8 +523,10 @@ class CSR(_csr_base):
             assert self.ncols == other.nrows
 
         K = get_kernel()
-        with releasing(K.to_handle(self), K) as a_h:
-            with releasing(K.to_handle(other), K) as b_h:
+
+        # Helper for handling sharding
+        def mul(A, b_h):
+            with releasing(K.to_handle(A), K) as a_h:
                 if transpose:
                     c_h = K.mult_abt(a_h, b_h)
                 else:
@@ -471,7 +534,18 @@ class CSR(_csr_base):
                 with releasing(c_h, K):
                     crepr = K.from_handle(c_h)
 
-        return crepr
+            return crepr
+
+        if self.nnz <= K.max_nnz:
+            # Common / fast path - one matrix
+            with releasing(K.to_handle(other), K) as b_h:
+                return mul(self, b_h)
+        else:
+            # Too large, let's go sharding
+            shards = self._shard_rows(K.max_nnz)
+            with releasing(K.to_handle(other), K) as b_h:
+                sparts = [mul(s, b_h) for s in shards]
+            return CSR._assemble_shards(sparts)
 
     def mult_vec(self, v):
         """
@@ -485,8 +559,69 @@ class CSR(_csr_base):
         """
         assert v.shape == (self.ncols,)
         K = get_kernel()
-        with releasing(K.to_handle(self), K) as h:
-            return K.mult_vec(h, v)
+        if self.nnz <= K.max_nnz:
+            with releasing(K.to_handle(self), K) as h:
+                return K.mult_vec(h, v)
+        else:
+            shards = self._shard_rows(K.max_nnz)
+            svs = []
+            for s in shards:
+                with releasing(K.to_handle(s), K) as h:
+                    svs.append(K.mult_vec(h, v))
+            return np.concatenate(svs)
+
+    def _shard_rows(self, tgt_nnz):
+        """
+        Shard a matrix by rows to fit in a target size.
+        """
+        assert tgt_nnz > 0
+
+        rest = self
+        shards = []
+        while rest.nnz > tgt_nnz:
+            # find the first split point
+            split = np.searchsorted(rest.rowptrs, tgt_nnz)
+            # if the start of the found row is too large, back up by one
+            if rest.rowptrs[split] > tgt_nnz:
+                if split <= 1:
+                    raise ValueError("row too large to fit in target matrix size")
+                split -= 1
+
+            _log.debug('splitting %s at %d (rp@s: %d)', rest, split, rest.rowptrs[split])
+            shards.append(rest.subset_rows(0, split))
+            rest = rest.subset_rows(split, rest.nrows)
+
+        shards.append(rest)
+        return shards
+
+    @classmethod
+    def _assemble_shards(cls, shards):
+        """
+        Reassemble a matrix from sharded rows.
+        """
+        nrows = sum(s.nrows for s in shards)
+        ncols = max(s.ncols for s in shards)
+        nnz = sum(s.nnz for s in shards)
+
+        rps = np.zeros(nrows + 1, np.int64)
+        rs = 0
+        for s in shards:
+            off = rps[rs]
+            re = rs + s.nrows + 1
+            rps[rs:re] = s.rowptrs + off
+            rs += s.nrows
+
+        assert rps[nrows] == nnz, f'{rps[nrows]} != {nnz}'
+
+        cis = np.concatenate([s.colinds for s in shards])
+        assert len(cis) == nnz
+        if shards[0].values is not None:
+            vs = np.concatenate([s.values for s in shards])
+            assert len(vs) == nnz
+        else:
+            vs = None
+
+        return cls(nrows, ncols, nnz, rps, cis, vs)
 
     def drop_values(self):
         """
@@ -522,7 +657,7 @@ class CSR(_csr_base):
         repr += '  rowptrs={}\n'.format(self.rowptrs)
         repr += '  colinds={}\n'.format(self.colinds)
         repr += '  values={}\n'.format(self.values)
-        repr += '}'
+        repr += '}>'
         return repr
 
     def __reduce__(self):
